from typing import Any, Dict, List, Union

JSONSchema = Dict[str, Any]


class Promptbuf:
    def __init__(self, schema: JSONSchema):
        self.schema = schema

    def encode(self, value: Any) -> str:
        output = []

        self.__encode(v=value, s=self.schema, output=output)

        return " ".join(output)

    def __encode(self, v: Any, s: JSONSchema, output: List[str]):
        s_type = s.get("type")

        if s_type == "object":
            ov = []
            if not s.get("properties"):
                return "{}"

            p_keys = list(s["properties"].keys())
            for key in p_keys:
                self.__encode(v.get(key), s["properties"][key], ov)

            output.append("{" + " ".join(ov) + "}")

        elif s_type == "array":
            if not s.get("items"):
                return "[]"
            av = []

            for item in v:
                self.__encode(item, s["items"], av)

            output.append("[" + " ".join(av) + "]")

        elif s_type == "string":
            if s.get("enum"):
                output.append(str(s["enum"].index(v)))
            else:
                output.append(f'"{v}"')

        elif s_type in ["integer", "number"]:
            output.append(str(v))

        elif s_type == "boolean":
            output.append("1" if v else "0")

        elif s_type == "null":
            output.append("null")

        else:
            return ""

    def decode(self, value: str) -> Any:
        return self.__decode(v=list(value), s=self.schema)

    def __decode(self, v: List[str], s: JSONSchema) -> Any:
        if not v:
            return

        char = v.pop(0)

        if char == "{":
            return self.__decode_object(v, s)
        elif char == "[":
            return self.__decode_array(v, s)
        elif char == '"':
            return self.__decode_string(v)
        elif char in [" ", "}", "]"]:
            return self.__decode(v, s)
        else:
            v.insert(0, char)
            return self.__decode_primitive(v, s)

    def __decode_object(self, v: List[str], s: JSONSchema) -> Dict[str, Any]:
        o = {}

        if not s.get("properties"):
            return None

        p_keys = list(s["properties"].keys())

        for key in p_keys:
            o[key] = self.__decode(v, s["properties"][key])

        return o

    def __decode_array(self, v: List[str], s: JSONSchema) -> List[Any]:
        a = []

        while len(v) > 0 and v[0] != "]":
            value = self.__decode(v, s["items"])

            if value:
                a.append(value)

        return a

    def __decode_string(self, v: List[str]) -> str:
        value = self.read_until(v, ['"'])
        v.pop(0)  # Remove the closing quote
        return value

    def __decode_primitive(
        self, v: List[str], s: JSONSchema
    ) -> Union[int, float, bool, None, str]:
        value = self.read_until(v, [" ", "}", "]"])

        s_type = s.get("type")

        if s_type == "integer":
            return int(value)
        elif s_type == "number":
            return float(value)
        elif s_type == "boolean":
            return value == "1"
        elif s_type == "null":
            return None
        elif s_type == "string":
            if s.get("enum"):
                return s["enum"][int(value)]
            return value

    def read_until(self, v: List[str], delimiters: List[str]) -> str:
        value = ""

        while v and v[0] not in delimiters:
            value += v.pop(0)

        return value



import json
from typing import Any, Dict
from enum import Enum
import pytest
from promptbuf import Promptbuf

JSONSchema = Dict[str, Any]


class Color(Enum):
    RED = "red"
    GREEN = "green"
    BLUE = "blue"


def run_test(name: str, schema: JSONSchema, value: Any):
    pb = Promptbuf(schema)
    encoded = pb.encode(value)
    decoded = pb.decode(encoded)
    assert json.dumps(value) == json.dumps(decoded), f"Test failed: {name}"


@pytest.mark.parametrize(
    "name,schema,value",
    [
        ("Integer, single digit", {"type": "integer"}, 1),
        ("Integer, many digits", {"type": "integer"}, 15),
        ("Number, single digit", {"type": "number"}, 2.5),
        ("Number, many digit", {"type": "number"}, 2307.588),
        ("String", {"type": "string"}, "Test string"),
        ("Boolean", {"type": "boolean"}, True),
        ("Null", {"type": "null"}, None),
        (
            "Enum",
            {"type": "string", "enum": ["red", "green", "blue"]},
            Color.BLUE.value,
        ),
    ],
)
def test_primitives(name, schema, value):
    run_test(name, schema, value)


@pytest.mark.parametrize(
    "name,schema,value",
    [
        (
            "Array of integers, single digit",
            {"type": "array", "items": {"type": "integer"}},
            [1, 2],
        ),
        (
            "Array of integers, many digits",
            {"type": "array", "items": {"type": "integer"}},
            [11, 22],
        ),
        (
            "Array of numbers, single digit",
            {"type": "array", "items": {"type": "number"}},
            [1.5, 2.5],
        ),
        (
            "Array of numbers, many digits",
            {"type": "array", "items": {"type": "number"}},
            [11.55, 22.55],
        ),
        (
            "Array of strings",
            {"type": "array", "items": {"type": "string"}},
            ["test", "string"],
        ),
        (
            "Array of enums",
            {
                "type": "array",
                "items": {"type": "string", "enum": ["red", "green", "blue"]},
            },
            [Color.BLUE.value, Color.GREEN.value, Color.RED.value],
        ),
    ],
)
def test_arrays(name, schema, value):
    run_test(name, schema, value)


@pytest.mark.parametrize(
    "name,schema,value",
    [
        (
            "Simple object",
            {
                "type": "object",
                "properties": {
                    "name": {"type": "string"},
                    "age": {"type": "integer"},
                    "isEnabled": {"type": "boolean"},
                },
            },
            {"name": "Tyler", "age": 29, "isEnabled": True},
        ),
        (
            "Simple object with array",
            {
                "type": "object",
                "properties": {
                    "name": {"type": "string"},
                    "age": {"type": "integer"},
                    "isEnabled": {"type": "boolean"},
                    "ids": {"type": "array", "items": {"type": "integer"}},
                },
            },
            {"name": "Tyler", "age": 29, "isEnabled": False, "ids": [1, 2, 3, 4, 5]},
        ),
        (
            "Nested objects",
            {
                "type": "object",
                "properties": {
                    "age": {"type": "integer"},
                    "isEnabled": {"type": "boolean"},
                },
            },
            {"age": 29, "isEnabled": True},
        ),
        (
            "Nested objects with name",
            {
                "type": "object",
                "properties": {
                    "name": {
                        "type": "object",
                        "properties": {
                            "first": {"type": "string"},
                            "last": {"type": "string"},
                        },
                    },
                },
            },
            {"name": {"first": "Tyler", "last": "O'Briant"}},
        ),
        (
            "Complex nested object",
            {
                "type": "object",
                "properties": {
                    "name": {
                        "type": "object",
                        "properties": {
                            "first": {"type": "string"},
                            "last": {"type": "string"},
                        },
                    },
                    "age": {"type": "integer"},
                    "isEnabled": {"type": "boolean"},
                },
            },
            {
                "name": {"first": "Tyler", "last": "O'Briant"},
                "age": 29,
                "isEnabled": True,
            },
        ),
        (
            "Object with array and primitives",
            {
                "type": "object",
                "properties": {
                    "ids": {
                        "type": "array",
                        "items": {"type": "integer"},
                    },
                    "age": {"type": "integer"},
                    "isEnabled": {"type": "boolean"},
                },
            },
            {"ids": [1, 2, 3, 4, 5], "age": 29, "isEnabled": False},
        ),
        (
            "Nested array of objects",
            {
                "type": "object",
                "properties": {
                    "employees": {
                        "type": "array",
                        "items": {
                            "type": "object",
                            "properties": {
                                "name": {
                                    "type": "string",
                                },
                                "role": {
                                    "type": "string",
                                },
                                "salary": {
                                    "type": "integer",
                                },
                            },
                        },
                    },
                },
            },
            {
                "employees": [
                    {
                        "name": "Alice",
                        "role": "Developer",
                        "salary": 70000,
                    },
                    {
                        "name": "Bob",
                        "role": "Designer",
                        "salary": 60000,
                    },
                ],
            },
        ),
    ],
)
def test_objects(name, schema, value):
    run_test(name, schema, value)



from setuptools import setup, find_packages

setup(
    name="promptbuf",
    version="0.1.1",
    author="Tyler O'Briant",
    author_email="tyler@tetraresearch.io",
    description="A brief description of promptbuf",
    long_description=open("README.md").read(),
    long_description_content_type="text/markdown",
    url="https://github.com/Tetra-Research/promptbuf",
    package_dir={"": "src"},
    packages=find_packages(where="src"),
    classifiers=[],
    python_requires=">=3.6",
    install_requires=[],
)
